import greenfoot.Greenfoot;
import greenfoot.GreenfootImage;

/**
 * Representation of a baddie (polluter). The baddie is loosing waste. If the
 * boy enters the cell with the baddie he will be displaced.
 * 
 * @author Dietrich Boles (University of Oldenburg, Germany)
 * @version 01.12.2009
 */
public class Baddie extends Item {

	public final static int RIGHT = 0;
	public final static int DOWN = 1;
	public final static int LEFT = 2;
	public final static int UP = 3;

	protected int col;
	protected int row;
	protected int oldCol;
	protected int oldRow;
	protected int direction;
	protected WasteWorld world;
	private Class<?>[] wasteClasses;

	private int number;
	protected int numberOfPollutions;
	protected int maxNumberOfPollutions;
	protected int cycleDuration;

	private GreenfootImage[] imgs;

	/**
	 * @param c
	 *            the column of the baddie
	 * @param r
	 *            the row of the baddie
	 * @param d
	 *            the direction of the baddie
	 * @param cls
	 *            class objects of the waste classes
	 */
	public Baddie(int c, int r, int d, Class<?>... cls) {
		this.imgs = new GreenfootImage[4];
		for (int i = 0; i < 4; i++) {
			this.imgs[i] = new GreenfootImage("baddie-" + i + "-0" + ".png");
		}

		this.world = WasteWorld.getWasteWorld();
		this.col = c;
		this.row = r;
		this.setDirection(d);
		this.wasteClasses = cls;
		this.number = 0;
		this.maxNumberOfPollutions = 20;
		this.numberOfPollutions = 0;
		this.cycleDuration = 120;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see Item#isAccessible(WasteWorld, int, int)
	 */
	@Override
	public boolean isAccessible(WasteWorld world, int col, int row) {
		return true;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see Item#draw(greenfoot.GreenfootImage, int, int, int)
	 */
	@Override
	public void draw(GreenfootImage img, int c, int r, int cellSize) {
		img.drawImage(this.getImage(), (c + 1) * cellSize, (r + 1) * cellSize);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see Item#handleCollisionBeforeMove(WasteWorld, int, int)
	 */
	@Override
	public void handleCollisionBeforeMove(WasteWorld world, int col, int row) {
		// displace the baddie
		Sounds.kill();
		world.removeItem(col, row);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see Item#handleCollisionAfterScrolling(WasteWorld, int, int)
	 */
	@Override
	public void handleCollisionAfterScrolling(WasteWorld world, int col, int row) {
		// displace the baddie
		Sounds.kill();
		world.removeItem(col, row);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see Item#act(WasteWorld, int, int)
	 */
	@Override
	public void act(WasteWorld world, int col, int row) {
		this.number++;
		this.oldCol = col;
		this.oldRow = row;
		if (this.number % 10 == 0) {
			this.move();
			if (this.number >= this.cycleDuration) {
				this.pollute(world);
				this.number = 0;
			}
		}
	}

	GreenfootImage getImage(int direction) {
		return this.imgs[direction];
	}

	int getCol() {
		return this.col;
	}

	int getRow() {
		return this.row;
	}

	int getDirection() {
		return this.direction;
	}

	void setDirection(int d) {
		this.direction = d;
		this.setImage(this.imgs[this.direction]);
	}

	void turnRight() {
		this.direction = (this.direction + 1) % 4;
		this.setImage(this.imgs[this.direction]);
	}

	boolean freeAhead() {
		switch (this.direction) {
		case RIGHT:
			return !this.world.cellOccupied(this.col + 1, this.row);
		case DOWN:
			return !this.world.cellOccupied(this.col, this.row + 1);
		case LEFT:
			return !this.world.cellOccupied(this.col - 1, this.row);
		default:
			return !this.world.cellOccupied(this.col, this.row - 1);
		}
	}

	void moveAhead() {
		switch (this.direction) {
		case RIGHT:
			this.col = this.col + 1;
			break;
		case DOWN:
			this.row = this.row + 1;
			break;
		case LEFT:
			this.col = this.col - 1;
			break;
		default:
			this.row = this.row - 1;
			break;
		}
	}

	void move() {
		boolean moved = false;
		for (int i = 0; i < 4; i++) {
			if (this.freeAhead()) {
				this.moveAhead();
				moved = true;
				break;
			}
			this.turnRight();
		}
		if (moved) {
			this.world.removeItem(this.oldCol, this.oldRow);
			this.world.changeItem(this.col, this.row, this);
		}

	}

	protected void pollute(WasteWorld world) {
		if (this.numberOfPollutions < this.maxNumberOfPollutions
				&& !this.world.cellOccupied(this.oldCol, this.oldRow)) {
			int index = Greenfoot.getRandomNumber(this.wasteClasses.length);
			Class<Waste> cls = (Class<Waste>) this.wasteClasses[index];
			try {
				// new piece of waste
				Waste waste = cls.newInstance();
				this.world.checkAndChangeItem(this.oldCol, this.oldRow, waste);
				Rest.getRest().incrRest(1);
				this.numberOfPollutions++;
			} catch (Exception exc) {

			}
		}
	}

}
