import greenfoot.Actor;
import greenfoot.GreenfootImage;

import java.awt.Color;
import java.awt.Font;
import java.util.Calendar;

/**
 * Representation of a clock. Singleton pattern.
 * 
 * @author Dietrich Boles (University of Oldenburg, Germany)
 * @version 01.12.2009
 */
public class Clock extends Actor {
	final static int SIZE = 75;
	private GreenfootImage image;
	private GreenfootImage clockImage;
	private int maxTime;
	private int imageTime;
	private boolean suspended;
	private boolean stopped;
	private long startTime;
	private long pauseStartTime;
	private int pauseDuration;

	private static Clock clock = new Clock();

	public static Clock getClock() {
		return Clock.clock;
	}

	/**
	 * gets the clock
	 * 
	 * @param secs
	 *            start time
	 * @return the clock (singleton)
	 */
	public static Clock getClock(int secs) {
		Clock.clock.maxTime = secs;
		Clock.clock.updateImage();
		return Clock.clock;
	}

	private Clock() { // singleton
		this.suspended = false;
		this.stopped = false;
		this.maxTime = 0;
		this.imageTime = 0;
		this.startTime = 0;
		this.pauseStartTime = 0;
		this.pauseDuration = 0;
		this.clockImage = new GreenfootImage("clock.png");
		this.image = new GreenfootImage(this.clockImage.getWidth(),
				this.clockImage.getHeight());
		this.image.setTransparency(200);
		this.image.setColor(Color.black);
		this.image.setFont(new Font("Monospaced", Font.BOLD, 18));
		this.updateImage();
		this.setImage(this.image);
	}

	/**
	 * sets the time
	 * 
	 * @param secs
	 *            number of seconds
	 */
	public void setTime(int secs) {
		this.reset(secs);
		this.updateImage();
	}

	/**
	 * gets the remaining time
	 * 
	 * @return the remaining time
	 */
	public int getRestTime() {
		return this.imageTime;
	}

	/**
	 * starts the clock
	 * 
	 * @param sec
	 *            start time
	 */
	public void start(int sec) {
		this.reset(sec);
		this.startTime = Calendar.getInstance().getTimeInMillis();
		this.imageTime = this.maxTime;
		this.updateImage();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see greenfoot.Actor#act()
	 */
	@Override
	public void act() {
		if (this.suspended || this.stopped) {
			return;
		}

		long currentTime = Calendar.getInstance().getTimeInMillis();
		this.imageTime = this.maxTime
				- ((int) (currentTime - this.startTime) - this.pauseDuration)
				/ 1000;
		if (this.imageTime < 0) {
			this.imageTime = 0;
		}
		this.updateImage();
		if (this.imageTime == 0) { // time has expired
			((WasteWorld) Clock.clock.getWorld()).setGameOver(false);
		}
	}

	/**
	 * resume the clock
	 */
	public void resume() {
		this.suspended = false;
		long currentTime = Calendar.getInstance().getTimeInMillis();
		this.pauseDuration += (int) (currentTime - this.pauseStartTime);
	}

	/**
	 * suspend the clock
	 */
	public void suspend() {
		this.suspended = true;
		this.pauseStartTime = Calendar.getInstance().getTimeInMillis();
	}

	/**
	 * @return true, if the clock has been suspended
	 */
	public boolean isSuspended() {
		return this.suspended;
	}

	/**
	 * stop the clock
	 */
	public void stop() {
		this.stopped = true;
	}

	/**
	 * reset the clock
	 * 
	 * @param sec
	 *            start time
	 */
	public void reset(int sec) {
		this.suspended = false;
		this.stopped = false;
		this.maxTime = sec;
		this.pauseStartTime = 0;
		this.pauseDuration = 0;
		this.updateImage();
	}

	/**
	 * draw the clock image
	 */
	private void updateImage() {
		this.image.clear();
		this.image.drawImage(this.clockImage, 0, 0);
		String timeString = "" + this.imageTime;
		int slen = Metrics.stringWidth(timeString, this.image.getFont());
		this.image.drawString(timeString, 2 + Clock.SIZE / 2 - slen / 2, 55);
	}
}
