import greenfoot.Actor;
import greenfoot.GreenfootImage;

/**
 * representation of a level
 * 
 * @author Dietrich Boles (University of Oldenburg, Germany)
 * @version 01.12.2009
 */
public abstract class Level extends Actor {

	final public static int LEVELS = 8; // current number of levels

	private static Level[] levels = new Level[Level.LEVELS + 1];

	// level completed?
	protected boolean hasBeenFinished;

	// the item in the selection level corresponding to this level
	protected SelectionItem selectionItem;

	// the number of pieces of waste in this level
	protected int numberOfWaste;

	// A world filter can be assigned to a level. The world will be drawn
	// regarding the filter.
	protected Filter worldFilter;

	// A cell filter can be assigned to a level. The cells of the world will be
	// drawn regarding the filter.
	protected Filter cellFilter;

	// A global filter can be assigned to a level. The world will be drawn
	// regarding the filter (stronger as the world filter)
	protected Filter globalFilter;

	// static methods

	/**
	 * Initializes all levels.
	 * 
	 * @param cells
	 *            cells which shall be filled
	 * @return the selection level
	 */
	public static Level init(Item[][] cells) {
		// create the level objects
		Level.levels[0] = new LevelSelection();
		Level.levels[1] = new Level1();
		Level.levels[2] = new Level2();
		Level.levels[3] = new Level3();
		Level.levels[4] = new Level4();
		Level.levels[5] = new Level5();
		Level.levels[6] = new Level6();
		Level.levels[7] = new Level7();
		Level.levels[8] = new Level8();

		return Level.createAndGetLevel(0, cells); // start with the selection
		// level
	}

	/**
	 * checks whether there are levels not yet completed
	 * 
	 * @return true if there are levels not yet completed
	 */
	public static boolean existsNotFinishedLevel() {
		for (int l = 0; l < Level.LEVELS; l++) {
			if (!Level.levels[l + 1].hasBeenFinished()) {
				return true;
			}
		}
		return false;
	}

	/**
	 * return the level
	 * 
	 * @param number
	 *            number of the level to be returned
	 * @return the level with the corresponding number
	 */
	public static Level getLevel(int number) {
		return Level.levels[number];
	}

	/**
	 * returns the selection level
	 * 
	 * @param cells
	 *            cells which shall be filled
	 * @return the selection level
	 */
	public static Level getSelectionLevel(Item[][] cells) {
		Level level = Level.levels[0];
		level.createLevel(cells);
		return level;
	}

	/**
	 * creates and return a certain level
	 * 
	 * @param number
	 *            number of the level
	 * @param cells
	 *            cells which shall be filled
	 * @return the level with the corresponding number
	 */
	public static Level createAndGetLevel(int number, Item[][] cells) {
		if (number < 0) {
			number = 0;
		} else if (number > Level.LEVELS) {
			number = Level.LEVELS;
		}
		Level level = Level.levels[number];
		level.createLevel(cells);
		return level;
	}

	// non-static methods

	protected Level(String img, SelectionItem selItem) {
		this.hasBeenFinished = false;
		this.setImage(img);
		this.selectionItem = selItem;
		this.numberOfWaste = 0;
		this.worldFilter = null;
		this.cellFilter = null;
		this.globalFilter = null;
	}

	/**
	 * level has been finished
	 */
	public void setFinished() {
		this.hasBeenFinished = true;
	}

	/**
	 * checks whether the level has been completed
	 * 
	 * @return true if the level has been completed
	 */
	public boolean hasBeenFinished() {
		return this.hasBeenFinished;
	}

	/**
	 * A world filter can be assigned to a level. The world will be drawn
	 * regarding the filter.
	 * 
	 * @return the world filter for this level (or null)
	 */
	public Filter getWorldFilter() {
		return this.worldFilter;
	}

	/**
	 * A cell filter can be assigned to a level. The cells of the world will be
	 * drawn regarding the filter.
	 * 
	 * @return the cell filter for this level (or null)
	 */
	public Filter getCellFilter() {
		return this.cellFilter;
	}

	/**
	 * A global filter can be assigned to a level. The world will be drawn
	 * regarding the filter (stronger as a world filter).
	 * 
	 * @return the global filter for this level (or null)
	 */
	public Filter getGlobalFilter() {
		return this.globalFilter;
	}

	/**
	 * return the number of pieces of waste in this level
	 * 
	 * @return the number of pieces of waste in this level
	 */
	public int getNumberOfWaste() {
		return this.numberOfWaste;
	}

	/**
	 * return the item in the selection level corresponding to this level.
	 * 
	 * @return the item in the selection level corresponding to this level
	 */
	public SelectionItem getSelectionItem() {
		return this.selectionItem;
	}

	/**
	 * return the position of a free cell
	 * 
	 * @param items
	 *            the cells
	 * @return the position of a free cell
	 */
	protected Position getFreeCell(Item[][] items) {
		return WasteWorld.getWasteWorld().getFreeCell();
	}

	/**
	 * creates a level
	 * 
	 * @param cells
	 *            the cells which shall be filled
	 */
	public abstract void createLevel(Item[][] cells);

	/**
	 * returns the time in which this level has to be completed
	 * 
	 * @return the time in which this level has to be completed
	 */
	public abstract int getSeconds();

	/**
	 * draws hints for this level
	 * 
	 * @param image
	 *            the image into which the hints shall be drawn
	 */
	public abstract void drawHelp(GreenfootImage image);

}
