import greenfoot.Actor;
import greenfoot.GreenfootImage;

import java.util.ArrayList;

/**
 * Representation of the storage (singleton pattern). The storage contains items
 * which the boy picked (waste, key, ladder, ...)
 * 
 * @author Dietrich Boles (University of Oldenburg, Germany)
 * @version 01.12.2009
 */
public class Storage extends Actor {
	final static int MAX_NUMBER = 3;

	private static Storage storage = new Storage();

	public static Storage getStorage() {
		return Storage.storage;
	}

	private Item[] items;
	private int number;
	private GreenfootImage img, storageImage;

	private Storage() { // singleton
		this.items = new Item[Storage.MAX_NUMBER];
		for (int i = 0; i < Storage.MAX_NUMBER; i++) {
			this.items[i] = null;
		}
		this.number = 0;
		this.storageImage = new GreenfootImage("storage.png");
		this.img = new GreenfootImage("storage.png");
		this.img.setTransparency(200);
		this.setImage(this.img);
	}

	/**
	 * checks whether the storage is empty
	 * 
	 * @return true if the storage is empty
	 */
	public boolean isEmpty() {
		return this.number == 0;
	}

	/**
	 * checks whether the storage is full
	 * 
	 * @return true if the storage is full
	 */
	public boolean isFull() {
		return this.number == Storage.MAX_NUMBER;
	}

	/**
	 * adds an item to the storage
	 * 
	 * @param i
	 *            the item which is added to the storage
	 */
	public void addItem(Item i) {
		if (this.number < Storage.MAX_NUMBER) {
			this.items[this.number++] = i;
			this.updateImage();
		}
	}

	/**
	 * checks whether the storage contains a curtain item
	 * 
	 * @param type
	 *            see the constants declared in class Item
	 * @return true, if the storage contains a curtain item
	 */
	public boolean containsItem(int type) {
		for (int i = 0; i < this.number; i++) {
			if (this.items[i].getType() == type) {
				return true;
			}
		}
		return false;
	}

	/**
	 * return an item of the storage of a curtain type
	 * 
	 * @param type
	 *            see the constants declared in class Item
	 * @return an item of a curtain type (or null)
	 */
	public Item getItem(int type) {
		for (int i = 0; i < this.number; i++) {
			if (this.items[i].getType() == type) {
				return this.items[i];
			}
		}
		return null;
	}

	/**
	 * checks whether the storage contains a piece of waste
	 * 
	 * @return true, if the storage contains a piece of waste
	 */
	public boolean containsWaste() {
		for (int i = 0; i < this.number; i++) {
			if (this.items[i].getType() >= Item.MIN_WASTE_TYPE
					&& this.items[i].getType() <= Item.MAX_WASTE_TYPE) {
				return true;
			}
		}
		return false;
	}

	/**
	 * return and removes all item of the storage of a curtain type
	 * 
	 * @param type
	 *            see the constants declared in class Item
	 * @return all item of the storage of a curtain type
	 */
	public Item[] removeItems(int type) {
		ArrayList<Item> remItems = new ArrayList<Item>(this.number);
		for (int i = 0; i < this.number; i++) {
			if (this.items[i].getType() == type) {
				remItems.add(this.items[i]);
				this.items[i] = null;
			}
		}
		Item[] res = remItems.toArray(new Item[0]);
		Item[] newItems = new Item[Storage.MAX_NUMBER];
		int j = 0;
		for (int i = 0; i < this.number; i++) {
			if (this.items[i] != null) {
				newItems[j++] = this.items[i];
			}
		}
		this.items = newItems;
		this.number = this.number - res.length;
		this.updateImage();

		return res;
	}

	/**
	 * return and removes one item of the storage of a curtain type
	 * 
	 * @param type
	 *            see the constants declared in class Item
	 * @return one item of the storage of a curtain type (or null)
	 */
	public Item removeOneItem(int type) {
		Item resItem = null;
		for (int i = 0; i < this.number; i++) {
			if (this.items[i].getType() == type) {
				resItem = this.items[i];
				this.items[i] = null;
				break;
			}
		}
		if (resItem != null) {
			Item[] newItems = new Item[Storage.MAX_NUMBER];
			int j = 0;
			for (int i = 0; i < this.number; i++) {
				if (this.items[i] != null) {
					newItems[j++] = this.items[i];
				}
			}
			this.items = newItems;
			this.number = this.number - 1;
			this.updateImage();
		}

		return resItem;
	}

	/**
	 * clears the storage
	 */
	public void clear() {
		for (int i = 0; i < Storage.MAX_NUMBER; i++) {
			this.items[i] = null;
		}
		this.number = 0;
		this.updateImage();
	}

	private void updateImage() {
		this.img.drawImage(this.storageImage, 0, 0);
		for (int i = 0; i < this.number; i++) {
			this.img.drawImage(this.items[i].getImage(), (i + 1) * 10 + i
					* WasteWorld.CELLSIZE, 10);
		}
	}
}
