import greenfoot.Greenfoot;
import greenfoot.GreenfootImage;
import greenfoot.World;

import java.awt.Color;

/**
 * Representation of the waste world.
 * 
 * @author Dietrich Boles (University of Oldenburg, Germany)
 * @version 01.12.2009
 */
public class WasteWorld extends World {

	// number of rows and cols
	public final static int ROWS = 25;
	public final static int COLS = 35;
	public final static int CELLSIZE = 36;

	// the visible part of the world
	public final static int HEIGHT = 13 * WasteWorld.CELLSIZE;
	public final static int WIDTH = 15 * WasteWorld.CELLSIZE;

	private static WasteWorld wasteWorld;

	// actors and items
	private Item[][] items;
	private Boy boy;
	private Map map;
	private Help help;
	private Storage storage;
	private Clock clock;
	private Rest rest;
	private Points points;
	private Level level;

	// helper variables
	private boolean mapVisible;
	private boolean helpVisible;
	private int scrollDiff;
	private int scrollDir;
	private boolean gameOver;
	private boolean stopped;
	private boolean finished;
	private boolean resetted;
	private boolean nextLevel;
	private boolean newLevel;

	private GreenfootImage fullImage, worldImage;

	public static WasteWorld getWasteWorld() {
		return WasteWorld.wasteWorld;
	}

	public WasteWorld() {
		super(WasteWorld.WIDTH, WasteWorld.HEIGHT, 1);
		WasteWorld.wasteWorld = this;
		this.fullImage = new GreenfootImage((WasteWorld.COLS + 2)
				* WasteWorld.CELLSIZE, (WasteWorld.ROWS + 2)
				* WasteWorld.CELLSIZE);
		this.worldImage = new GreenfootImage(WasteWorld.WIDTH,
				WasteWorld.HEIGHT);

		// create empty cells
		this.items = new Item[WasteWorld.COLS][WasteWorld.ROWS];
		for (int c = 0; c < WasteWorld.COLS; c++) {
			for (int r = 0; r < WasteWorld.ROWS; r++) {
				this.items[c][r] = null;
			}
		}

		// create the levels
		this.level = Level.init(this.items);
		this.addObject(this.level, this.level.getImage().getWidth() / 2, 80);

		// create the boy
		this.boy = new Boy(this, 0, 0, Boy.DOWN);
		this.addObject(this.boy, WasteWorld.WIDTH / 2, WasteWorld.HEIGHT / 2);

		// create the clock
		this.clock = Clock.getClock(this.level.getSeconds());
		this.addObject(this.clock, this.clock.getImage().getWidth() / 2,
				WasteWorld.HEIGHT - this.clock.getImage().getHeight() / 2);

		// create the storage; the boy can pick items which will be stored in
		// the storage
		this.storage = Storage.getStorage();
		this.storage.clear();
		this.addObject(this.storage, this.clock.getImage().getWidth() + 10
				+ this.storage.getImage().getWidth() / 2, WasteWorld.HEIGHT
				- this.storage.getImage().getHeight() / 2);

		// rest: indicates how many pieces of waste are in the world yet
		this.rest = Rest.getRest();
		this.rest.setRest(this.level.getNumberOfWaste());
		this.addObject(this.rest, this.clock.getImage().getWidth() + 10
				+ this.storage.getImage().getWidth() + 10
				+ this.rest.getImage().getWidth() / 2, WasteWorld.HEIGHT
				- this.rest.getImage().getHeight() / 2);

		// points: aim of the game is to get as much points as possible
		this.points = Points.getPoints();
		this.points.clear();
		this.addObject(this.points, WasteWorld.WIDTH
				- this.points.getImage().getWidth() / 2, WasteWorld.HEIGHT
				- this.points.getImage().getHeight() / 2);

		this.setBackground(this.createBackgroundImage(0, Boy.DOWN));

		Greenfoot.setSpeed(41);

		this.staticInit();

		// helper variables
		this.map = null;
		this.help = null;
		this.mapVisible = false;
		this.helpVisible = false;
		this.scrollDiff = -1;
		this.scrollDir = Boy.DOWN;
		this.resetted = true;
		this.stopped = false;
		this.gameOver = false;
		this.finished = false;
		this.nextLevel = false;
		this.newLevel = false;
	}

	void staticInit() {
		Fire.staticInit();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see greenfoot.World#started()
	 */
	@Override
	public void started() {
		if (this.gameOver) {
			this.stopped = true;
			Greenfoot.stop();
			return;
		}

		if (!this.resetted && this.clock.isSuspended()) {
			this.clock.resume();
			return;
		}

		Sounds.startLevel();
		this.resetted = false;
		this.removeObject(this.level);
		this.clock.start(this.level.getSeconds());
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see greenfoot.World#stopped()
	 */
	@Override
	public void stopped() {
		this.clock.suspend();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see greenfoot.World#act()
	 */
	@Override
	public void act() {
		if (this.stopped) {
			return;
		}

		if (this.scrollDiff > 0) {
			// implements the smooth moving of the boy
			this.checkScrollBackground();
			return;
		}

		if (this.gameOver) { // game over
			this.handleGameOver();
			return;
		}

		if (this.nextLevel) { // start selection level
			this.clearKeyBuffer();
			this.startSelectionLevel();
			this.nextLevel = false;
			return;
		}

		if (this.newLevel) { // start new level
			this.clearKeyBuffer();
			this.newLevel = false;
			return;
		}

		this.checkMovingKeys(); // moving the boy

		if (this.nextLevel || this.newLevel || this.gameOver) {
			this.newLevel = false;
			return;
		}

		this.checkMapKey(); // showing a small map

		this.checkHelp(); // showing hints

		this.itemsAct(); // explicitly calling the act-method of the items of
		// the world
	}

	public Boy getBoy() {
		return this.boy;
	}

	public Level getLevel() {
		return this.level;
	}

	public Item[][] getItems() {
		return this.items;
	}

	public void removeItem(int col, int row) {
		this.items[col][row] = null;
	}

	public void changeItem(int col, int row, Item item) {
		this.items[col][row] = item;
	}

	public boolean checkAndChangeItem(int col, int row, Item item) {
		if (col >= 0 && col < WasteWorld.COLS && row >= 0
				&& row < WasteWorld.ROWS && this.items[col][row] == null
				&& !(this.boy.getCol() == col && this.boy.getRow() == row)) {
			this.items[col][row] = item;
			return true;
		}
		return false;
	}

	public boolean cellAccessible(int col, int row) {
		return col >= 0
				&& col < WasteWorld.COLS
				&& row >= 0
				&& row < WasteWorld.ROWS
				&& (this.items[col][row] == null || this.items[col][row]
						.isAccessible(this, col, row));
	}

	public boolean cellOccupied(int col, int row) {
		return col < 0 || col >= WasteWorld.COLS || row < 0
				|| row >= WasteWorld.ROWS || this.items[col][row] != null
				|| this.boy.getCol() == col && this.boy.getRow() == row;
	}

	protected Position getFreeCell() {
		Position pos = new Position();
		do {
			pos.setCol(Greenfoot.getRandomNumber(WasteWorld.COLS));
			pos.setRow(Greenfoot.getRandomNumber(WasteWorld.ROWS));
		} while (this.cellOccupied(pos.getCol(), pos.getRow()));
		return pos;
	}

	private void startSelectionLevel() {
		this.handleTimeBonus();
		this.startLevel(Level.getSelectionLevel(this.items));
		this.clock.start(this.level.getSeconds());
	}

	private void startLevel(Level l) {
		this.level = l;
		this.scrollDiff = 0;
		this.storage.clear();
		this.rest.setRest(this.level.getNumberOfWaste());
		this.boy.reset(0, 0, Boy.DOWN);
		this.setBackground(this.createBackgroundImage(0, Boy.DOWN));
		this.clock.setTime(this.level.getSeconds());
		this.addObject(this.level, this.level.getImage().getWidth() / 2, 80);
		this.repaint();
		Sounds.startLevel();
		Greenfoot.delay(20);
		this.removeObject(this.level);
	}

	// called by SelectionItem
	public void setSelectedLevel(Level l) {
		this.clock.stop();
		this.newLevel = true;
		this.startLevel(l);
		this.clock.start(l.getSeconds());
	}

	// called by Dustbin
	public void setLevelFinished() {
		this.clock.stop();
		this.nextLevel = true;
	}

	private void handleTimeBonus() {
		Greenfoot.delay(3);
		Sounds.endLevel();
		int bonusPoints = this.clock.getRestTime();
		EndLevel endLevel = new EndLevel();
		this.addObject(endLevel, WasteWorld.WIDTH / 2, endLevel.getImage()
				.getHeight() / 2 + 2);
		TimeBonus bonus = new TimeBonus(bonusPoints);
		this.addObject(bonus, this.boy.getX(), this.boy.getY()
				- this.boy.getImage().getHeight());
		this.repaint();
		Greenfoot.delay(40);
		this.points.incrPoints(bonusPoints);
		this.removeObject(endLevel);
		this.removeObject(bonus);
	}

	// called by Clock (time limit exceeded) and Dustbin (no more waste in the
	// world)
	public void setGameOver(boolean lastLevel) {
		this.clock.stop();
		this.gameOver = true;
		this.finished = lastLevel;
	}

	private void handleGameOver() {
		if (this.finished) {
			this.handleTimeBonus();
			EndFinished end = new EndFinished(this.points.getPointsNumber());
			this.addObject(end, WasteWorld.WIDTH / 2, WasteWorld.HEIGHT / 2);
			Sounds.finished();
			Greenfoot.delay(50);
		} else {
			EndTimeProblem end = new EndTimeProblem(this.points
					.getPointsNumber());
			this.addObject(end, WasteWorld.WIDTH / 2, WasteWorld.HEIGHT / 2);
			Sounds.timeLimitExceeded();
			Greenfoot.delay(50);
		}
		Greenfoot.stop();
	}

	private void itemsAct() {
		for (int c = 0; c < WasteWorld.COLS; c++) {
			for (int r = 0; r < WasteWorld.ROWS; r++) {
				if (this.items[c][r] != null) {
					// calling the act-method explicitly
					// note: items are not added to the world!
					this.items[c][r].act(this, c, r);
				}
			}
		}
	}

	// implementing the smooth movement of the boy
	private void checkScrollBackground() {
		this.scrollDiff -= WasteWorld.CELLSIZE / 3;
		this.setBackground(this.createBackgroundImage(this.scrollDiff,
				this.scrollDir));

		if (this.scrollDiff <= 0) {
			int c = this.boy.getCol();
			int r = this.boy.getRow();
			Item item = this.items[c][r];
			if (item != null) {
				// the boy entered an occupied cell
				item.handleCollisionAfterScrolling(this, c, r);
			}
		}
	}

	private void clearKeyBuffer() {
		// there seems to be a bug in greenfoot
		Greenfoot.isKeyDown("right");
		Greenfoot.isKeyDown("left");
		Greenfoot.isKeyDown("up");
		Greenfoot.isKeyDown("down");
	}

	private void checkMovingKeys() {

		if (this.boy.isSick()) {
			return;
		}

		int check = 0;

		boolean right = Greenfoot.isKeyDown("right");
		boolean left = Greenfoot.isKeyDown("left");
		boolean up = Greenfoot.isKeyDown("up");
		boolean down = Greenfoot.isKeyDown("down");

		if (right) {
			check++;
		}
		if (left) {
			check++;
		}
		if (up) {
			check++;
		}
		if (down) {
			check++;
		}

		if (check >= 2) {
			return; // only one key/direction
		}

		boolean ok = false;

		int oldCol = this.boy.getCol();
		int oldRow = this.boy.getRow();

		if (right) {
			ok = this.boy.right();
			this.scrollDir = Boy.RIGHT;
		} else if (left) {
			ok = this.boy.left();
			this.scrollDir = Boy.LEFT;
		} else if (up) {
			ok = this.boy.up();
			this.scrollDir = Boy.UP;
		} else if (down) {
			ok = this.boy.down();
			this.scrollDir = Boy.DOWN;
		}

		if (ok) {
			Item item = this.items[oldCol][oldRow];
			if (item != null) {
				// the boy leaves an occupied cell
				item.handleCollisionBeforeMove(this, oldCol, oldRow);
			}

			int c = this.boy.getCol();
			int r = this.boy.getRow();
			item = this.items[c][r];
			if (item != null) {
				// the boy will enter an occupied cell
				item.handleCollision(this, c, r);
			}

			this.scrollDiff = 2 * WasteWorld.CELLSIZE / 3;
			this.setBackground(this.createBackgroundImage(this.scrollDiff,
					this.scrollDir));
		} else {
			this.setBackground(this.createBackgroundImage(0, this.scrollDir));
		}
	}

	private void checkMapKey() {
		if (!this.mapVisible && Greenfoot.isKeyDown("m")) {

			// a small map of the world will be shown
			this.map = new Map(this.createMap());
			this.addObject(this.map, WasteWorld.WIDTH
					- this.map.getImage().getWidth() / 2, WasteWorld.HEIGHT
					- this.map.getImage().getHeight() / 2);
			this.mapVisible = true;
		} else if (this.mapVisible && !Greenfoot.isKeyDown("m")) {

			// the map will be invisible
			this.removeObject(this.map);
			this.mapVisible = false;
		}
	}

	private void checkHelp() {
		if (!this.helpVisible && Greenfoot.isKeyDown("h")) {

			// some hints will be shown
			this.help = new Help(this.level);
			this.addObject(this.help, WasteWorld.WIDTH / 2,
					WasteWorld.HEIGHT / 2);
			this.helpVisible = true;
		} else if (this.helpVisible && !Greenfoot.isKeyDown("h")) {

			// the hints will be removed
			this.removeObject(this.help);
			this.helpVisible = false;
		}
	}

	// creates an image of the whole landscape (all cells)
	private GreenfootImage createFullImage(GreenfootImage img, boolean withBoy) {

		img.clear();

		// frame
		for (int c = 0; c < WasteWorld.COLS + 2; c++) {
			img.drawImage(new Wall().getImage(), c * WasteWorld.CELLSIZE, 0);
			img.drawImage(new Wall().getImage(), c * WasteWorld.CELLSIZE,
					(WasteWorld.ROWS + 1) * WasteWorld.CELLSIZE);
		}
		for (int r = 1; r <= WasteWorld.ROWS; r++) {
			img.drawImage(new Wall().getImage(), 0, r * WasteWorld.CELLSIZE);
			img.drawImage(new Wall().getImage(), (WasteWorld.COLS + 1)
					* WasteWorld.CELLSIZE, r * WasteWorld.CELLSIZE);
		}

		// cells
		for (int c = 0; c < WasteWorld.COLS; c++) {
			for (int r = 0; r < WasteWorld.ROWS; r++) {
				img.drawImage(Cell.getCellImage(), (c + 1)
						* WasteWorld.CELLSIZE, (r + 1) * WasteWorld.CELLSIZE);
				Filter filter = this.level.getCellFilter();
				if (filter != null) {
					img.drawImage(filter.getImage(), (c + 1)
							* WasteWorld.CELLSIZE, (r + 1)
							* WasteWorld.CELLSIZE);
				}
				if (this.items[c][r] != null) {
					this.items[c][r].draw(img, c, r, WasteWorld.CELLSIZE);
				}
			}
		}

		if (withBoy) {
			GreenfootImage cImg = this.boy.getImage(this.boy.getDirection());
			img.drawImage(cImg, (this.boy.getCol() + 1) * WasteWorld.CELLSIZE
					+ (WasteWorld.CELLSIZE - cImg.getWidth()) / 2, (this.boy
					.getRow() + 1)
					* WasteWorld.CELLSIZE
					+ (WasteWorld.CELLSIZE - cImg.getHeight()) / 2);
		}

		return img;
	}

	// creates an image of the visible part of the landscape
	private GreenfootImage createBackgroundImage(int diff, int dir) {
		this.worldImage.clear();
		this.worldImage.setColor(new Color(153, 204, 0));
		this.worldImage.fill();
		int x = (this.boy.getCol() + 1) * WasteWorld.CELLSIZE;
		int y = (this.boy.getRow() + 1) * WasteWorld.CELLSIZE;
		this.createFullImage(this.fullImage, false);
		switch (dir) {
		case Boy.RIGHT:
			this.worldImage.drawImage(this.fullImage, WasteWorld.WIDTH / 2 - x
					- WasteWorld.CELLSIZE / 2 + diff, WasteWorld.HEIGHT / 2 - y
					- WasteWorld.CELLSIZE / 2);
			break;
		case Boy.LEFT:
			this.worldImage.drawImage(this.fullImage, WasteWorld.WIDTH / 2 - x
					- WasteWorld.CELLSIZE / 2 - diff, WasteWorld.HEIGHT / 2 - y
					- WasteWorld.CELLSIZE / 2);
			break;
		case Boy.DOWN:
			this.worldImage.drawImage(this.fullImage, WasteWorld.WIDTH / 2 - x
					- WasteWorld.CELLSIZE / 2, WasteWorld.HEIGHT / 2 - y
					- WasteWorld.CELLSIZE / 2 + diff);
			break;
		default:
			this.worldImage.drawImage(this.fullImage, WasteWorld.WIDTH / 2 - x
					- WasteWorld.CELLSIZE / 2, WasteWorld.HEIGHT / 2 - y
					- WasteWorld.CELLSIZE / 2 - diff);
			break;
		}

		// filtering

		Filter filter = this.level.getWorldFilter();
		if (filter != null) {
			this.worldImage.drawImage(filter.getImage(), 0, 0);
		}
		filter = this.level.getGlobalFilter();
		if (filter != null) {
			this.worldImage.drawImage(filter.getImage(), 0, 0);
		}

		this.boy.setDiff(diff / (WasteWorld.CELLSIZE / 3));

		return this.worldImage;
	}

	// creates a small map of the landscape
	private GreenfootImage createMap() {
		GreenfootImage mapImage = new GreenfootImage((WasteWorld.COLS + 2)
				* WasteWorld.CELLSIZE, (WasteWorld.ROWS + 2)
				* WasteWorld.CELLSIZE);
		this.createFullImage(mapImage, true);
		mapImage.scale(mapImage.getWidth() / 3, mapImage.getHeight() / 3);
		return mapImage;
	}

}
